/* filter_evdev/main.c
  
  Created: 2009-11
  Updated: 2009-11-26
  
  Main file of the filter_evdev program.

  This is free software licensed under the WTFPL

            DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE 
                    Version 3, November 2009 

 Copyright (C) 2009 André Gillibert
  La vieille église, 14370 Vimont, France
 Everyone is permitted to copy and distribute verbatim or modified 
 copies of this license document, and changing it is allowed as long 
 as the name is changed. 

            DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE 
   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION 

  0. You just DO WHAT THE FUCK YOU WANT TO.
*/

#define _ISOC99_SOURCE
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <fcntl.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <ctype.h>
#include <math.h>
#include "filter_evdev.h"

#undef DEBUG_YYLEX
#undef DEBUG_GRAMMAR
#undef DEBUG_FILTERS
#undef DEBUG_EVALUATION
#undef DEBUG_READSETTINGS
#undef DEBUG_ACCL

static int devin;  /* file descriptor to physical input device (evdev) */
static int devout; /* file descriptor to virtual input device (uinput) */

typedef unsigned char byte;

/* returns 1 if bit number i is set, otherwise returns 0 */
int bit_set(size_t i, const byte* a)
{
  return a[i/CHAR_BIT] & (1 << i%CHAR_BIT);
}

/* Assign features (supported axes and keys) of the physical input device (devin)
 * to the virtual input device (devout) */
static void copy_features(void)
{
  byte evtypes[EV_MAX/CHAR_BIT + 1] = {0};
  byte codes[KEY_MAX/CHAR_BIT + 1];
  unsigned i,code;
  int op;
  if (ioctl(devin, EVIOCGBIT(0, sizeof(evtypes)), evtypes) < 0) return;
  for(i=0;i<EV_MAX;++i)
    if (bit_set(i, evtypes)) {
      switch(i) {
        case EV_KEY: op = UI_SET_KEYBIT; break;
	case EV_REL: op = UI_SET_RELBIT; break;
	case EV_ABS: op = UI_SET_ABSBIT; break;
	case EV_MSC: op = UI_SET_MSCBIT; break;
	case EV_LED: op = UI_SET_LEDBIT; break;
	case EV_SND: op = UI_SET_SNDBIT; break;
	case EV_SW: op = UI_SET_SWBIT; break;
	default: op = -1;
      }
      if (op == -1) continue;
      ioctl(devout, UI_SET_EVBIT, i);
      memset(codes,0,sizeof(codes));
      if (ioctl(devin, EVIOCGBIT(i, sizeof(codes)), codes) < 0) return;
      for(code=0;code<KEY_MAX;code++)
        if (bit_set(code, codes)) ioctl(devout, op, code);
    }
}

/* The filter structure can contain
 * 1) F_RANGE: A range condition. Specific value range e.g. [-3,+5].
 * 2) F_CODE: A type/code condition. e.g. EV_REL/REL_X.
 * 3) F_OPERATE: An operation: Multiplication, addition or comparison. e.g. * 3.
 * 4) F_MAPTO: A mapping command: e.g. map EV_REL/REL_X
 * 5) F_UNMAP: An unmap indicator.
 * 6) F_END_INSTRUCTION: An end-of-instruction mark.
 * 7) F_COMMIT: A commit operation.
 * 
 * filters_head -> ... -> filters_queue is a singly linked lists representing
 * the current set of recorded commands to apply to every event set.
 *
 * A translation command is represented as a sequence of F_CODE, F_RANGE, 
 * F_OPERATE, F_MAPTO and F_UNMAP (more than one F_UNMAP 
 * is equivalent to one F_UNMAP) filters and ended with a F_END_INSTRUCTION filter.
 * These filter structs may be in any order.
 *
 * A commit command is represented as a single F_COMMIT command.
 */
struct filter
{
	struct filter *next; /* always defined (NULL for end) */
	enum filter_type {
		  F_RANGE		/* u.range must be defined           */
		, F_CODE		/* u.code must be defined            */
		, F_END_INSTRUCTION	/* no field defined but filter::next */
		, F_COMMIT		/* no field defined but filter::next */
		, F_MAPTO		/* mapto defined                     */
		, F_UNMAP		/* no field defined byt filter::next */
		, F_OPERATE		/* u.op defined                      */
	} type;
	union filter_params {
		struct range {
			double min, max;
		} range;
		struct code code;
		struct operate {
			enum operator_type
				{
				OP_MULT
				, OP_ADD
				, OP_LOWER   /* yields 1 if value < operand, 0 otherwise */
				, OP_GREATER
				}
			operator;
			double operand;
		} op;
		struct code mapto;
	} u;
} *filters_head, *filters_queue; /* Set of recorded commands */

#ifdef DEBUG_FILTERS
static void display_filter(const struct filter *f)
{
	fprintf(stderr, "(filter ");
	if (f->type == F_END_INSTRUCTION) {
		fprintf(stderr, "end sequence");
	} else if (f->type == F_COMMIT) {
		fprintf(stderr, "commit");
	}
	if (f->type == F_UNMAP) {
		fprintf(stderr, "unmap");
	} else if (f->type == F_MAPTO) {
		fprintf(stderr, "mapto %u/%u", f->u.mapto.type, f->u.mapto.code);
	} else if (f->type == F_OPERATE) {
		struct operate op = f->u.op;
		fprintf(stderr, "%c %f",
			op.operator == OP_MULT ? '*'
			: op.operator == OP_ADD ? '+'
			: op.operator == OP_GREATER ? '>'
			: op.operator == OP_LOWER ? '<'
			: '?'
			,op.operand);
	} else if (f->type == F_RANGE) {
		struct range r = f->u.range;
		fprintf(stderr, "[%g, %g]", r.min, r.max);
	} else if (f->type == F_CODE) {
		fprintf(stderr, "%u/%u", f->u.code.type, f->u.code.code);
	}
	fprintf(stderr,")");
}
static void display_filters(void)
{
	struct filter *f=filters_head;
	while(f) {
		display_filter(f);
		f=f->next;
	}
	fprintf(stderr, "\n");
}
#endif

/* Create a new filter structure whose all fields but filter::next are copied from f
 * and appends it to the recorded set of commands filters_head->filters_queue.
 * It becomes the new filters_queue 
 * Returns a pointer to the newly created filter struct */
struct filter *new_filter(struct filter f)
{
	struct filter *nf = malloc(sizeof(struct filter));
	if (!nf) return NULL;
	*nf = f;
	nf->next = NULL;
	if (!filters_head) {
		filters_head = filters_queue = nf;
	} else {
		filters_queue->next = nf;
		filters_queue = nf;
	}
	return nf;
}
/* Erase all filter structs following FILTER, but not FILTER itself, 
 * in filters_head->filters_queue list.
 * Erase everything if FILTER is null */
static void erase_filters_after(struct filter *filter)
{
	if (filter) {
		filters_queue = filter;
		filter = filter->next;
	} else {
		filter = filters_head;
		filters_queue = filters_head = NULL;
	}
	while(filter) {
		struct filter *next=filter->next;
		free(filter);
		filter=next;
	}
}

/* A pointer to the current character on the currently parsed command string
 * This variable is set by parse_command_line() and read/incremented by yylex() */
static const char *tokline;

/* A token value: Either a float value or a type/code value (e.g. EV_REL/REL_X) 
 * The effective type is defined by the yylex() function return value.
 * This is similar to the lex parsing system */
static union yylval_type
{
	double fval;
	struct code code;
} yylval;

/* A token type, as returned by yylex() */

enum toktype
{
	 FLOAT = 261  /* An integer or FP value. yylval.fval is the value. */
	,CLEAR = 262  /* The "clear" token */
	,MAPTO = 263  /* The "map" token */
	,UNMAP = 264  /* The "unmap" token */
	,COMMIT = 265 /* The "commit" token */
	,EVENT_CODE = 266 /* A type/code value such as EV_REL/REL_X.
			   yylval.code is the value */
};
/* Store a type/code value in *c, for a predefined string s
 * Such as REL_X or BTN_LEFT.
 * Returns 1 on success and 0 if the string is not found. */
static int codeof(const char *s, struct code *c)
{
	const struct code_name *p=code_names;
	for (;p->name;p++) {
		if (strcmp(p->name, s)==0) {
			*c = p->code;
			return 1;
		}
	}
	return 0;
}
/* Extract a token from tokline, and update the tokline so that
 * it points to the next token.
 * Assign the token value to yylval.
 * Returns the token type (toktype). 
 * Special tokens [ ] + - * , < > are returned as their ASCII value.
 * Returns zero on error or when tokens are exhausted (end of line met).
 */
static int yylex(void)
{
	int ln;
	int c;
	char *s,*end;
	unsigned code=-1;
	char *sm;
#ifdef DEBUG_YYLEX
	fprintf(stderr, "[next token]");
#endif
	
	/* ignore leading whitespace */
	while(*tokline == ' ' || *tokline == '\t') tokline++;
	if (!*tokline) return 0;
	c=*tokline;
	if (c == 'X') code = REL_X;
	else if (c == 'Y') code = REL_Y;
	else if (c == 'Z') code = REL_Z;
	if (code != -1) { /* X, Y and Z tokens are EVENT_CODEs */
		yylval.code.type = EV_REL;
		yylval.code.code = code;
		tokline++;
		return EVENT_CODE;
	}
	if (c == '[' || c == ']' || c == ',' || c == '*' || c == '+'
		|| c == '<' || c == '>'
		|| (c=='-' && !isdigit(tokline[1]) && tokline[1] != '.')) {
#ifdef DEBUG_YYLEX
		fprintf(stderr, "[token %c]", *tokline);
#endif
		return *tokline++;
	}
	ln=0;
	/* It's a multi-character token. Search for whitespace or delimiter */
	while(c != ' ' && c != '\t' && c != '[' && c != ']'
		&& c !=',' && c != '*' && c != '\0')
		{c=*++tokline;ln++;}
	if (ln==0) return 0; /* defensive programming */
	s=malloc(ln+1); /* s will contain a nul-terminated string representing the word */
	if (!s) return 0;
	memcpy(s, tokline-ln, ln);
	s[ln]=0;
#ifdef DEBUG_YYLEX
	fprintf(stderr, "[token string %s]", s);
#endif
	if (strcmp(s, "clear")==0) return CLEAR;
	else if (strcmp(s, "map")==0) return MAPTO;
	else if (strcmp(s, "unmap")==0) return UNMAP;
	else if (strcmp(s, "remap")==0) {tokline -= 3;return UNMAP;} /* funny hack */
	else if (strcmp(s, "commit")==0) return COMMIT;
	else if (codeof(s, &yylval.code)) return EVENT_CODE;
	/* Is it "type/code" ? */
	if ((sm=strchr(s, '/'))) {
		*sm = '\0';
		yylval.code.type = strtol(s, &end, 10);
		if (end == sm) {
			yylval.code.code = strtol(sm+1, &end, 10);
			if (end == s+ln) return EVENT_CODE;
		}
		*sm = '/';
	}
	/* Is it a integer/float? */
	yylval.fval=strtod(s,&end);
#ifdef DEBUG_YYLEX
	fprintf(stderr, "[token float %f]", yylval.fval);
#endif
	free(s);
	if (end == s+ln) return FLOAT;
#ifdef DEBUG_YYLEX
	fprintf(stderr, "[invalid token]");
#endif
	return 0;
}

/* Read tokens on the current parsing line (for commands passed through a socket)
 * get_range reads tokens of a range with this syntax:
 * (+|-)? [ <float>? , <float>? ]
 * 
 * lex is a the last token that has been read (it's usually an open bracket).
 * get_range returns the token that follows the closing bracket.
 * yylval is set to the related value.
 *
 * get_range has side effects:
 * Any range found is appended to the filter list (filters_head->filters_queue).
 * If the range sign is unspecified, both the negative and positive ranges
 * are inserted as two distinct filter structs in the list.
 *
 * If no range is found, it's interpreted as a [-inf, +inf] range.
 * In that case, no token is eaten.
 */
static int get_range(int lex)
{
	/* back up char position to restore it in case of failure */
	const char *otokline = tokline; 
	/* and back up the previous token with its value too */
	union yylval_type yy = yylval;
	int olex = lex;
	double rmin, rmax; /* min and max values of the range interval */
	/* range_sign is 1 if range is prefixed with a +,
	 * -1 if it's prefixed with a -, 
	 * and zero if it has no prefix */
	int range_sign; 
	struct filter f={0};
	f.type = F_RANGE;

	/***** Get range sign *****/
	if (lex == '+') range_sign = 1;
	else if (lex == '-') range_sign = -1;
	else range_sign = 0;
	if (range_sign) lex=yylex();
#ifdef DEBUG_GRAMMAR
	fprintf(stderr, "[range_sign %d]", range_sign);
#endif

	/***** Get range bounds *****/
	if (lex != '[') goto fail;
	lex = yylex();
	if (lex == ',') rmin = -INFINITY;
	else if (lex == FLOAT) {rmin = yylval.fval;lex=yylex();}
	else goto fail;
#ifdef DEBUG_GRAMMAR
	fprintf(stderr, "[rmin %f]", rmin);
#endif
	if (lex != ',') goto fail;
	lex = yylex();
	if (lex == ']') rmax = INFINITY;
	else if (lex == FLOAT) {rmax = yylval.fval;lex=yylex();}
	else goto fail;
#ifdef DEBUG_GRAMMAR
	fprintf(stderr, "[rmax %f]", rmax);
#endif
	if (lex != ']') goto fail;
	/**** Create filter structs for this range *****/
	if (range_sign >= 0) {
		f.u.range.min = rmin;
		f.u.range.max = rmax;
		if (!new_filter(f)) goto fail;
	}
	if (range_sign <=0) {
		f.u.range.min = -rmax;
		f.u.range.max = -rmin;
		if (!new_filter(f)) goto fail;
	}
	return yylex();

	fail:
	yylval = yy;
	tokline = otokline;
	f.u.range.min = -INFINITY;
	f.u.range.max = INFINITY;
	new_filter(f);
	return olex;
}
/* Similar in interface to get_range.
 * Input: lex/yylval of the first token.
 * Output: Return lex value of the token following the last axis/key code
 * and assign yylval to the related value.
 *
 * Tokens syntax that it consumes:
 * <event code>*
 * event code: <symbolic constant>|<type>/<code>
 * type: <integer> such as 2 for EV_REL.
 * code: <integer> such as 1 for REL_Y.
 * symbolic constant: One of REL_X, BTN_LEFT, etc.
 *
 * Side effects: Appends code conditions to the
 * filter list (filters_head->filters_queue).
 * It appends one filter struct for each type/code.
 */
static int get_axis_codes(int lex)
{
	struct filter f={0};
	
	f.type = F_CODE;
	f.u.code.type = EV_REL;
	while (lex == EVENT_CODE) {
		f.u.code = yylval.code;
		if (!new_filter(f)) break;
		lex = yylex();
	}
	return lex;
}
/* parse_command_line parses and records an instruction line received from
 * the configuration socket.
 * It appends the commands to the filter list (filters_head->filters_queue).
 *
 * A translation command is recorded as a set of conditions and remapping filters
 * ended with a F_END_INSTRUCTION filter struct.
 *
 * A commit command is recorded as a single F_COMMIT filter struct.
 *
 * A clear command is immediately executed: It resets the whole filter list!
 * Return value: 1 on success, 0 on syntax error or OOM condition.
 * Examples setting_line:
 * "XY [0, 5] * 2"
 * "2/0 +[8,12] - 3 remap REL_Y map REL_Z"
 * "commit"
 * "clear"
 */
static int parse_command_line(const char *setting_line)
{
	/* setting_line: <translation command>|commit|clear
	 * translation command: <event code>* <range>? <operation>*
	 * event code: <integer>|X|Y|Z
	 * range: (+|-)? [ <float>?, <float>? ]
	 * operation: <arithmetic operation>|<remapping operation>
	 * arithmetic operation: (+|-|*|<|>) <float>
	 * remapping operation: unmap | (map|remap) <event code>
	 */
	/* Back up current queue of the filter list so that we
	 * can easily cancel changes in case of 
	 * setting_line syntax error or OOM condition */
	struct filter *filter_base = filters_queue;
	struct filter f={0};
	int lex;
	int r; /* function's return value. i.e. 1 on success */

#ifdef DEBUG_GRAMMAR
	fprintf(stderr, "parse_command_line %s\n", setting_line);
#endif

	tokline = setting_line;
	
	lex = yylex();
	if (lex == CLEAR) {
		erase_filters_after(NULL);
		return 1;
	} else if (lex == COMMIT) {
		f.type = F_COMMIT;
		return new_filter(f) != NULL;
	}
#ifdef DEBUG_GRAMMAR
	fprintf(stderr,"[searching axis codes at %s]", tokline);
#endif
	/***** Get codes *****/
	lex = get_axis_codes(lex);
	/***** Get ranges *****/
#ifdef DEBUG_GRAMMAR
	fprintf(stderr,"[searching range at %s]", tokline);
#endif
	lex = get_range(lex);

	/***** Get operations (arithmetic or remapping) *****/
	while (lex) {
		int op = OP_ADD;
		
#ifdef DEBUG_GRAMMAR
		fprintf(stderr,"[searching command at %s]", tokline);
#endif
		if (lex == '*' || lex == '<' || lex == '>') {
			op =	(lex == '*' ? OP_MULT
				:lex == '<' ? OP_LOWER
				:lex == '>' ? OP_GREATER
				:OP_ADD);
			lex = yylex();
		} else if (lex == '+') {
			lex = yylex();
		} else if (lex == '-') {
			lex = yylex();
			if (lex == FLOAT) yylval.fval = -yylval.fval;
		} /* else if lex == FLOAT do nothing */
		if (lex == FLOAT) {
			f.type = F_OPERATE;
			f.u.op.operator = op;
			f.u.op.operand = yylval.fval;
		} else if (lex == UNMAP) {
#ifdef DEBUG_GRAMMAR
			fprintf(stderr,"[unmap found]");
#endif
			f.type = F_UNMAP;
		} else if (lex == MAPTO) {
#ifdef DEBUG_GRAMMAR
			fprintf(stderr,"[map found]");
#endif
			f.type = F_MAPTO;
			lex = yylex();
			if (lex != EVENT_CODE) {
#ifdef DEBUG_GRAMMAR
				fprintf(stderr, "[expected axis code]");
#endif
				goto fail;
			}
			f.u.mapto = yylval.code;
		} else goto fail;
		if (!new_filter(f)) goto fail;
#ifdef DEBUG_GRAMMAR
		fprintf(stderr,"[next command?]");
#endif
		lex = yylex();
	}
	f.type = F_END_INSTRUCTION;
#ifdef DEBUG_GRAMMAR
	fprintf(stderr, "[line parsed]\n");
#endif
	r = (new_filter(f) != NULL);
#ifdef DEBUG_FILTERS
	display_filters();
#endif
	return r;
	
	fail:
	erase_filters_after(filter_base);
	return 0;
}
/* File descriptor of the server UNIX socket.
 * It's listened */
static int fd_settings=-1; 

/* Path to the server UNIX socket file
 * It's set from argv in main() */
static const char *settings_path;

/* Open UNIX socket file descriptor and configure
 * it so that we can select() and accept() on fd_settings */
static int open_settings(void)
{
	struct sockaddr_un addr;

	addr.sun_family = AF_UNIX;
	strncpy(addr.sun_path, settings_path, sizeof(addr.sun_path));
	addr.sun_path[sizeof(addr.sun_path)-1]=0;
	fd_settings = socket(AF_UNIX, SOCK_STREAM, 0);
	if (fd_settings == -1) return 0;
	unlink(settings_path);
	if (bind(fd_settings
		, (struct sockaddr*)&addr
		, sizeof(struct sockaddr_un))==-1) goto fail_clo;
	if (listen(fd_settings, 32)==-1) goto fail_clo;
	return 1;

	fail_clo:
	close(fd_settings);
	fd_settings=-1;
	return 0;
}

/* Reads, parses and records commands on a ifd client socket
 * It may contain more than one line.
 * Each line is individually parsed and recorded.
 */
static int record_client_request(int ifd)
{
	static char linebuf[256];
	static char *plinebuf=linebuf;
	ssize_t cb;

	plinebuf = linebuf;
	while((cb=read(ifd, plinebuf, linebuf+sizeof(linebuf)-1-plinebuf)) > 0) {
		char *p;
		plinebuf[cb] = 0;
		while ((p=strchr(plinebuf, '\n'))) {
			cb = cb-(p-plinebuf)-1;
			*p=0;
#ifdef DEBUG_READSETTINGS
			fprintf(stderr,"[new line %s]", linebuf);
#endif
			parse_command_line(linebuf);
			memmove(linebuf, p+1, cb);
			plinebuf = linebuf;
			plinebuf[cb]=0;
		}
		plinebuf += cb;
	}
	return 1;
}

/* Accept a connection on the server UNIX socket
 * and records commands sent from this client
 * connection.
 * Note that:
 * 1) The whole program is blocked until the client
 * session is ended. The client should send all its
 * commands at once, very quickly, and close the connection.
 *
 * 2) This function should be called when the server
 * is "ready" as returned by pselect(2). i.e. there's
 * a pending client connection that's going to be
 * immediately accepted.
 */
static int read_from_settings_fd(void)
{
	struct sockaddr_un addr;
	socklen_t len=sizeof(addr);
	int ifd;
	
	if ((ifd=accept(fd_settings, (struct sockaddr*)&addr, &len))==-1) {
		return 0;
	}
	record_client_request(ifd);
	shutdown(ifd, SHUT_RDWR);
	close(ifd);
	return 1;
}

/* struct finput_event represents an input event
 * with a floating point value.
 * ev.value should be ignored. fval is the real value
 */
struct finput_event
{
	struct input_event ev;
	double fval;
	struct finput_event *next;
};
/*
 * pevepack->...->pevpack_e is a singly linked list of finput_events
 * representing the current set of simultaneous events.
 *
 * pevnpack->...->pevnpack_e is a list of finput_events
 * representing the next state of the current set of simultaneous events
 * as it'll be when the sequence of translation commands down to the
 * next "commit" command, is applied.
 * That is: pevnpack is progressively built when executing commands.
 *
 * When a "commit" command is executed, it drops the current pevpack list and
 * replace it with the now-complete pevnpack list, and reset pevnpack to NULL
 * so that, it's ready to receive the state for the execution of the next
 * sequence of commands up to the next commit command or the end of the command set.
 */

struct finput_event *pevpack,*pevpack_e,*pevnpack,*pevnpack_e;/* sets of simultaneous events */

/* Appends a new struct fevent to the specified linked list
 * (This should be pevpack or pevnpack)
 * The contents of this new struct fevent are derived from ev
 */
static struct finput_event *new_fevent
	(struct finput_event **pevs, struct finput_event **peve
	, struct finput_event ev)
{

	struct finput_event *nev = malloc(sizeof(struct finput_event));
	if (!nev) return NULL;
	*nev = ev;
	nev->next = NULL;
	if (!*pevs) {
		*pevs = nev;
	} else (*peve)->next = nev;
	*peve = nev;
	return nev;
}
/* Create a new finput_event based on ev, and appends it to the
 * pevnpack list representing the next transformed state of the current set
 * of simultaneous events */
static struct finput_event *new_nfevent(struct finput_event ev)
{
	return new_fevent(&pevnpack, &pevnpack_e, ev);
}

/* Frees all nodes of a  linked list of finput_event structures */
static void clear_evlist(struct finput_event **ps, struct finput_event **pe)
{
	struct finput_event *f=*ps;

	while (f) {
		struct finput_event *next = f->next;
		free(f);
		f=next;
	}
	*ps = *pe = NULL;
}

/* translate integer values to float values in the set
 * of simultaneous events represented by the pevpack list
 * That is: Copy ev.value to fval for each struct */
static void evfloat_from_evint(void)
{
	struct finput_event *l=pevpack;
#ifdef DEBUG_EVALUATION
	fprintf(stderr, "[evset begin]");
#endif
	while(l) {
#ifdef DEBUG_EVALUATION
		fprintf(stderr, "[event %u/%u val %ld]"
			, l->ev.type, l->ev.code
			, (long)l->ev.value);
#endif
		l->fval = l->ev.value;
		l=l->next;
	}
#ifdef DEBUG_EVALUATION
	fprintf(stderr, "[evset end]\n");
#endif
}

/* Similar to evfloat_from_evint, but, that time, copy back
 * float values (fval) to integer values (ev.value)
 * That's the last step before the set of simultaneous events
 * is dumped to the virtual input device */
static void evint_from_evfloat(void)
{
	struct finput_event *l=pevpack;
#ifdef DEBUG_EVALUATION
	fprintf(stderr, "[fset begin]");
#endif
	while(l) {
#ifdef DEBUG_EVALUATION
		fprintf(stderr, "[evf %u/%u val %g]"
			, l->ev.type, l->ev.code
			, l->fval);
#endif
		l->ev.value = l->fval;
		l=l->next;
	}
#ifdef DEBUG_EVALUATION
	fprintf(stderr, "[fset end]\n\n");
#endif
}
/* Execute a commit command. It drops the current pevpack
 * set of events and replace it with the pevnpack
 * set of transformed events */
static void commit(void)
{
	clear_evlist(&pevpack, &pevpack_e);
	pevpack   = pevnpack;
	pevpack_e = pevnpack_e;
	pevnpack=pevnpack_e=NULL;
}

/* Writes the current (transformed) set of simultaneous events
 * as stored in pevpack->pevpack_e, to the virtual input device
 */
static void output_events(void)
{
	struct finput_event *l=pevpack;
	while(l) {
		write(devout, &l->ev, sizeof(l->ev));
		l=l->next;
	}
}

/* Returns the type/code pair from an input_event struct */
static struct code code_from_event(struct input_event ev)
{
	struct code c;
	c.type = ev.type;c.code = ev.code;
	return c;
}

/* Searchs an event of the specified type in the pevnpack list
 * Returns NULL if not found */
static struct finput_event *find_nevent(struct code ncode)
{
	struct finput_event *l=pevnpack;

	while(l) {
		if (l->ev.type == ncode.type && l->ev.code == ncode.code)
			return l;
		l=l->next;
	}
	return NULL;
}

/* Execute a mapping operation.
 * That is: Create an event in the pevnpack list of the specified ncode type.
 * If the event already exists, the value is summed up to the current value
 * so that, mapping two axes to REL_Z will make REL_Z virtually equal to the
 * sum of both axes. */
static int mapto(const struct finput_event *ev, struct code ncode, double effval)
{
	struct finput_event *nev = find_nevent(ncode);
	if (!nev) {
		nev = new_nfevent(*ev);
		if (!nev) return 0;
		nev->ev.type = ncode.type;
		nev->ev.code = ncode.code;
		nev->fval = 0;
	}
	nev->fval += effval;
	return 1;
}

/* Search all type and range conditions of the current translation command
 * and returns 1 if the ev event matches them, or 0 otherwise.
 * f points to the first filter struct of the translation command line
 * The translation command line is ended with an F_END_INSTRUCTION object
 */
static int eval_condition(const struct finput_event *ev, const struct filter *f)
{
	int conditions=3;

	while(conditions && f && f->type != F_END_INSTRUCTION) {
		if (f->type == F_RANGE
		 && f->u.range.min <= ev->fval
		 && f->u.range.max > ev->fval)
			conditions &= 2;
		if (f->type == F_CODE
		 && f->u.code.type == ev->ev.type
		 && f->u.code.code == ev->ev.code)
			conditions &= 1;
		f=f->next;
	}
	return !conditions;
}
/* Executes all operations of the specified translation command 
 * on the ev->fval value, and returns the computed value.
 * f points to the first filter struct of the translation command line
 * The translation command line is ended with an F_END_INSTRUCTION object
 * eval_operations do not check for conditions. It should be called after a 
 * successful eval_condition.
 */
static double eval_operations(const struct finput_event *ev, const struct filter *f)
{
	double baseval, effval;

	baseval = effval = ev->fval;
	while(f && f->type!=F_END_INSTRUCTION) { /* apply operators */
		if (f->type == F_OPERATE) {
		 	if (f->u.op.operator == OP_ADD)
				effval += f->u.op.operand;
			else if (f->u.op.operator == OP_MULT)
				effval *= f->u.op.operand;
			else if (f->u.op.operator == OP_LOWER)
				effval = (effval < f->u.op.operand);
			else if (f->u.op.operator == OP_GREATER)
				effval = (effval > f->u.op.operand);
		}
		f=f->next;
	}
	return effval;
}

/* Execute all remapping operations, including self-mapping if there's no unmap operation
 * in the specified translation command.
 * The translation command is given as a f pointer pointing to the first filter of the
 * translation command. See eval_condition and eval_operation.
 *
 * The value computed by eval_operation must be passed as effval.
 * eval_mapping will append new events to pevnpack as needed
 */
static void eval_mapping(const struct finput_event *ev, double effval, const struct filter *f)
{
	int unmap=0;
	while(f && f->type!=F_END_INSTRUCTION) { /* remaps */
		if (f->type == F_MAPTO)
			mapto(ev, f->u.mapto, effval);
		else if (f->type == F_UNMAP) unmap = 1;
		f=f->next;
	}
	if (!unmap) mapto(ev, code_from_event(ev->ev), effval);
}

/* Input: Pointer to the start of a filter sequence ended
 * with a filter struct of type TYPE. e.g. translation command ended with
 * a F_END_INSTRUCTION.
 * Output: Pointer to the start of the next sequence or NULL 
 */
static const struct filter *skipto_filter(const struct filter *f, int type)
{
	while (f && f->type != type) f=f->next;
	if (f && f->type == type) f=f->next;
	return f;
}

/* Test conditions and apply operations and remappings as needed for
 * an event ev from the pevpack list, and create the resulting transformed
 * events in the pevnpack list */
static void translate_event(const struct finput_event *ev, const struct filter *bf)
{
	const struct filter *f=bf;
	int mapped=0;

	/*if (ev->ev.type != EV_REL) {
		new_nfevent(*ev);
		return;
	}*/
	while (f && f->type != F_COMMIT) {
		if (eval_condition(ev, f)) {
			mapped=1;
			eval_mapping(ev, eval_operations(ev, f), f);
		}
		f = skipto_filter(f, F_END_INSTRUCTION);
	}
	if (!mapped) mapto(ev, code_from_event(ev->ev), ev->fval);
}

/* Transform a set of simultaneous events 
 * collected in the pevpack list, and output
 * their transformed version to the virtual input device
 */
static void flush_evpack(void)
{
	const struct filter *f;
	static struct finput_event *l;
	if (!pevpack) return;
	evfloat_from_evint();

	f = filters_head;
	while (f) {
		l = pevpack;
		while (l) {
			translate_event(l, f);
			l=l->next;
		}
		commit();
		f = skipto_filter(f, F_COMMIT);
	}
	evint_from_evfloat();
	output_events();
	clear_evlist(&pevpack, &pevpack_e);
}

/* Read ready events from the physical input device
 * and transform/output them to the virtual input device
 * when a EV_SYN indicating the end of a sequence
 * of simultaneous events is met.
 *
 * read_from_input collects events in the pevpack list
 * and flushes it only when EV_SYN is met.
 *
 * This function returns when there are no more events ready
 * i.e. a call to read would block
 *
 * This function should be called after a pselect(2) on devin.
 */
static void read_from_input(void)
{
	int cb;
	int attempt;
	struct finput_event fev={0};

	if ((cb=read(devin, &fev.ev, sizeof(fev.ev)))!=-1
	&& cb==sizeof(fev.ev)) {
		attempt = 8;
		fev.fval = fev.ev.value;
		while (attempt && !new_fevent(&pevpack, &pevpack_e, fev)) {
			flush_evpack();
			attempt--;
		}
		if (!attempt) /* cannot allocate memory; don't map device */
			write(devout, &fev.ev, sizeof(fev.ev));
		else if (fev.ev.type == EV_SYN) flush_evpack();
	}
}

/* Find which file descriptor: Either physical input device or UNIX
 * server socket is ready for input first
 * That's a simple select(2) operation
 * 
 * Returns S_ERROR on SIGINT/SIGTERM/SIGQUIT or any other
 * fatal error.
 */
static enum which_stream {S_INPUT, S_SETTINGS, S_ERROR} select_fd(void)
{
	/* return which fd to read from: input device or server socket */
	static int round_robin=0;
	int max;
	fd_set s;

	start:
	FD_ZERO(&s);
	if (fd_settings!=-1) FD_SET(fd_settings, &s);
	FD_SET(devin, &s);

	if (devin > fd_settings) max = devin;
	else max = fd_settings;
	if (select(max+1, &s, NULL, NULL, NULL)==-1) {
		return S_ERROR;
	}
	if (fd_settings!=-1 && FD_ISSET(devin, &s) && FD_ISSET(fd_settings, &s)) {
		round_robin=1-round_robin;
		return round_robin ? S_INPUT : S_SETTINGS;
	}
	if (FD_ISSET(devin, &s)) return S_INPUT;
	if (fd_settings!=-1 && FD_ISSET(fd_settings, &s)) return S_SETTINGS;
	goto start;
}

/* Read-Eval-Print-Loop
 * That is: Wait for input on physical input device
 * and UNIX server socket, and when input is received
 * treat it as needed.
 * physical input device -> transformation -> virtual input device
 * UNIX server socket -> client connection -> parse and record commands
 */
static void repl(void)
{
	enum which_stream stream;
	while((stream=select_fd())!=S_ERROR) {
		if (stream == S_SETTINGS) read_from_settings_fd();
		if (stream == S_INPUT) read_from_input();
	}
}

/* main() function.
 * Parse parameters, open file descriptors:
 * devin: file descriptor of the physical input device
 * devout: file descriptor of the virtual input device
 * fd_settings: file descriptor of the UNIX server socket
 * Then, Read-Eval-Print-Loop (repl)
 */
int main(int argc, char** argv)
{
  struct uinput_user_dev devinfo={0};
  if (argc!=3) {
    fprintf(stderr,"%s: Two arguments expected: <evdev file> <settings fifo>\n", argv[0]);
    return 1;
  }
  devin=open(argv[1], O_RDONLY|O_SYNC|O_NONBLOCK);
  if (devin<0) {
    fprintf(stderr,"%s: Unable to open input device: %s\n", argv[0], argv[1]);
    return 2;
  }
  settings_path = argv[2];
  if (!open_settings()) {
    fprintf(stderr,"%s: Unable to open or bind settings UNIX domain socket: %s\n"
    	, argv[0], settings_path);
    return 2;
  }
  devout=open("/dev/misc/uinput",O_WRONLY|O_NONBLOCK);
  if (devout<0) {
    fprintf(stderr,"Unable to open /dev/misc/uinput (char device 10:223).\nPossible causes: Device node inexistent or kernel not compiled with evdev user level driver support or permission denied.\n");
    perror("open(\"/dev/misc/uinput\")");
    return 3;
  }
  ioctl(devin, EVIOCGNAME(UINPUT_MAX_NAME_SIZE), devinfo.name);
  strncat(devinfo.name, "+sensitivity", UINPUT_MAX_NAME_SIZE-1);
  fprintf(stderr, "%s\n", devinfo.name);
  ioctl(devin, EVIOCGID, &devinfo.id);

  copy_features();
  write(devout,&devinfo,sizeof(devinfo));
  if (ioctl(devout,UI_DEV_CREATE)<0) {
    fprintf(stderr,"Unable to create input device with UI_DEV_CREATE\n");
    return 2;
  } else {
    fprintf(stderr,"Device created.\n");
  }
  repl();
  return 0;
}
